﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Windows;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Threading.Tasks;
using System.Windows.Media;
using System.Windows.Threading;
using Arction.Wpf.Charting;
using Arction.Wpf.Charting.SeriesXY;
using Arction.Wpf.Charting.Views.ViewXY;
using DevExpress.DirectX.Common.DirectWrite;
using System.Drawing;
using System.Drawing.Imaging;
using System.Linq;
using System.Text;
using Arction.Wpf.Charting.Annotations;
using Arction.Wpf.Charting.Axes;

namespace WpfTaskChart
{

    /// <summary>
    ///  Objects that configures and uses LightningChart, as instructed by client. 
    ///  Communication with the client is provided by a PipeConnector. 
    /// </summary>
    /// <remarks>
    /// 
    /// </remarks>
    public class LCRunner
    {
        /// <summary>
        ///  LightningChart to serve.
        /// </summary>
        private LightningChartUltimate _chart = null;
        private bool _cursorChanged;
        /// <summary>
        ///  State of the thread. 
        /// </summary>
        private bool _running = false;
        
        /// <summary>
        ///  Logging method. 
        /// </summary>
        private readonly Action<string> LOG;

        /// <summary>
        /// Tick (time) data generation has started. 
        /// </summary>
        private double _post = 0;

        /// <summary>
        /// Tick (time) the last sample was generated. 
        /// </summary>
        private double _pre = 0;

        private double _step = 1.0;

        private readonly int _width;
        private readonly int _height;

        private List<System.Drawing.Bitmap> _imageList;

        private ImageFormat _format;

        private int _index;

        /// <summary>
        ///  Main constructor for the LC runner. 
        /// </summary>
        /// <param name="lOG">Logging method. </param>
        public LCRunner(Action<string> lOG,  string title, int index, double pre , double post, int width, int height, double step, ImageFormat imageFormat)
        {
            LOG = lOG;
            Title = title;
            _pre = pre;
            _post = post;
            _width = width;
            _height = height;
            _step = step;
            _imageList = new List<Bitmap>();
            _format = imageFormat;
            _index = index;
        }

        public string Title { get; set; }

        public List<Bitmap> GenerateImages()
        {
            var op = Application.Current.Dispatcher.Invoke(new Func<List<Bitmap>>(() =>
            {
                ReInitChart();
                for (var idx = _pre; idx < _post; idx += _step)
                {
                    var currentImage = RenderNewFrame(idx);
                    if(currentImage != null)
                        _imageList.Add(currentImage);
                }
                return _imageList;
            }), DispatcherPriority.Background, null);
            //op.Wait();
            return (List<Bitmap>)op;
        }

        ///// <summary>
        /////  Async operation to start the LC operations STA thread. 
        /////  <para>Does nothing if the thread is not running.</para>
        ///// </summary>
        //public async Task<List<Bitmap>> Start()
        //{
        //    if (!_running)
        //    {
        //        LOG("Starting LC thread");
        //        _running = true;
        //        return await Task.Run(() =>
        //        {

        //            for (var idx = _pre; idx < _post; idx += _step)
        //            {
        //                RenderNewFrame(idx);
        //            }

        //            App.Current.Dispatcher.BeginInvoke(new Action(() =>
        //            {
        //                int index = 1;
        //                _imageList.ForEach(x =>
        //                {
        //                    x.Save($"{_chart.Title.Text}_{index}.png", ImageFormat.Png);
        //                    index++;
        //                });
        //            }));
        //            return _imageList;

        //        }, token);
        //    }

        //    return  null;
        //}



        /// <summary>
        ///  Reinitializes the chart. 
        ///  <para>
        ///  Can be called from any thread. The initalization happens on the staThread. 
        ///  </para>
        ///  <para>
        ///  Note that the old chart reference is not checked, but just overridden. This will cause memory leaks if old chart is 
        ///  not properly disposed first. 
        /// </para>
        /// </summary>
        private void ReInitChart()
        {
            _chart = new LightningChartUltimate(new RenderingSettings()
            {
                HeadlessMode = true
            });
            

            // Tell chart to begin block where no new frames are to be produced. 
            _chart.BeginUpdate();

                var titleColor = System.Windows.Media.Color.FromRgb(249, 202, 3);
                var gridColor = System.Windows.Media.Color.FromArgb(35, 255, 255, 255);

                // Configure colors for the chart.
                _chart.ChartBackground.Color = System.Windows.Media.Color.FromRgb(16, 16, 17);
                _chart.ChartBackground.GradientFill = GradientFill.Solid;
                _chart.ViewXY.GraphBackground.Color = System.Windows.Media.Color.FromArgb(255, 20, 20, 20);
                _chart.ViewXY.GraphBackground.GradientFill = GradientFill.Solid;
                _chart.Title.Text = Title;
                _chart.Title.Color = titleColor;

                foreach (var yAxis in _chart.ViewXY.YAxes)
                {
                    yAxis.AxisThickness = 1;
                    yAxis.Title.Color = titleColor;
                    yAxis.Title.MouseHighlight = MouseOverHighlight.None;
                    yAxis.MajorGrid.Color = gridColor;
                    yAxis.MajorGrid.Pattern = LinePattern.Solid;
                    yAxis.MinorDivTickStyle.Visible = false;
                }

                foreach (var xAxis in _chart.ViewXY.XAxes)
                {
                    xAxis.AxisThickness = 1;
                    xAxis.Title.Color = titleColor;
                    xAxis.Title.MouseHighlight = MouseOverHighlight.None;
                    xAxis.MajorGrid.Color = gridColor;
                    xAxis.MajorGrid.Pattern = LinePattern.Solid;
                    xAxis.MinorDivTickStyle.Visible = false;
                }

                // Configure legend.
                _chart.ViewXY.LegendBoxes[0].Visible = false;

                // Initilize rendering device manually. This is required on WPF as it would not do it automatically wihtout parent. 
                _chart.InitializeRenderingDevice(true);

                // Chart requires manually defined size, as it does not have a parent that would define it. 
                _chart.Width = _width;
                _chart.Height = _height;

                // Async rendering is currently unavailable on a headless mode. Use Sync.
                _chart.ChartRenderOptions.UpdateType = ChartUpdateTypes.Sync;
                
                // Register to events. ChartError will receive all the information on 
                // problmes etc. AfterRendering notifies us that we have new frame available. 
                // _chart.ChartMessage += OnChartMessage;
                _chart.AfterRendering += chart_AfterRendering;
                // We could use outputstream instead of just watching for new frames and taking exports. 
                //_chart.OutputStream = _chartStream;
                //_chart.OutputStream.BeginRead(_chartBytes, 0, 4096, testcallback, null);

                // Allow chart to produce new frames, as the initialization procedures are over. 

                PointLineSeries series = new PointLineSeries(_chart.ViewXY, _chart.ViewXY.XAxes[0], _chart.ViewXY.YAxes[0]);
                series.AddPoints(GeneratePoints(_index), true);
                _chart.ViewXY.PointLineSeries.Add(series);

                var mainCursor = new LineSeriesCursor(_chart.ViewXY, _chart.ViewXY.XAxes[0])
                {
                    Tag = "Cursor #1",
                    LineStyle =
                    {
                        Width = 2f,
                        Color = Colors.Red
                    },
                    SnapToPoints = true,
                    Visible = true,
                    ValueAtXAxis = 5,
                    TrackPoint =
                    {
                        Color1 = Colors.White
                    }
                };
                mainCursor.PositionChanged += cursor_PositionChanged;
                 _chart.ViewXY.LineSeriesCursors.Add(mainCursor);
                 AnnotationXY cursorValueDisplay = new AnnotationXY(_chart.ViewXY, _chart.ViewXY.XAxes[0], _chart.ViewXY.YAxes[0]);
                 cursorValueDisplay.Style = AnnotationStyle.RoundedCallout;
                 cursorValueDisplay.LocationCoordinateSystem = CoordinateSystem.RelativeCoordinatesToTarget;
                 cursorValueDisplay.LocationRelativeOffset.X = 162;
                 cursorValueDisplay.LocationRelativeOffset.Y = -88;
                 cursorValueDisplay.Sizing = AnnotationXYSizing.Automatic;
                 cursorValueDisplay.TextStyle.Font = new WpfFont("Lucida Console", 13f, false, false);
                 cursorValueDisplay.TextStyle.Color = Colors.Black;
                 cursorValueDisplay.Text = "";
                 cursorValueDisplay.Fill.Color = System.Windows.Media.Color.FromArgb(128, 0, 0, 0);
                 cursorValueDisplay.TextStyle.Color = Colors.White;
                 cursorValueDisplay.BorderLineStyle.Color = System.Windows.Media.Color.FromArgb(64, 255, 255, 255);
                 cursorValueDisplay.BorderLineStyle.Width = 1;
                 cursorValueDisplay.Fill.GradientFill = GradientFill.Solid;
                 cursorValueDisplay.TargetMoveByMouse = false;
                 cursorValueDisplay.Shadow.Visible = false;
                 cursorValueDisplay.Visible = false;
                 cursorValueDisplay.MouseInteraction = false;
                _chart.ViewXY.Annotations.Add(cursorValueDisplay);
                 _chart.ViewXY.ZoomToFit();
            
                _chart.EndUpdate();
         
        }

        private void cursor_PositionChanged(object sender, PositionChangedEventArgs e)
        {
            // Cancel the on-going rendering because the code below updates 
            // the year chart.
            var chart = (sender as LineSeriesCursor).OwnerView.OwnerChart;
            e.CancelRendering = true;
            UpdateCursorResult(chart);
        }

        private SeriesPoint[] GeneratePoints(int index)
        {
            var rv = new SeriesPoint[30];
            var idx = 0;
            if (index == 1)
            {
                for (int i = 0; i < 30; i++)
                {
                    if (i >= 0 && i <= 5)
                        rv[idx++] = new SeriesPoint(i, 0.1);
                    if (i > 5 && i <= 10)
                        rv[idx++] = new SeriesPoint(i, 8.99);
                    if (i > 10)
                        rv[idx++] = new SeriesPoint(i, 0.99);
                }
            }
            else
            {
                for (int i = 0; i < 30; i++)
                {
                    if (i >= 0 && i <= 10)
                        rv[idx++] = new SeriesPoint(i, 0.1);
                    if (i > 10 && i <= 15)
                        rv[idx++] = new SeriesPoint(i, 8.99);
                    if (i > 15)
                        rv[idx++] = new SeriesPoint(i, 0.99);
                }
            }
            return rv;
        }

        /// <summary>
        /// React to new frames rendered by the chart. 
        /// <para>Send the rendered frame using the communication channl. </para>
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        /// <remarks>
        ///  Data from the chart could be obtained using OutputStream as well. 
        ///  It would be (or at least it will be) faster, but it produced the raw image only. 
        ///  This is easier as we wish to have it capsuled in bmp format for easier presentation on the client side. 
        /// </remarks>
        private void chart_AfterRendering(object sender, AfterRenderingEventArgs e)
        {
            LOG("new frame rendered.");

            // Init straem to save the chart to.
            //MemoryStream stream = new MemoryStream();

            //// Export last rendered frame to stream. 
            //// Due to current limitation of LightningChart Headless mode 
            //// the exportation must be done right after frame is rendred. 
            //bool ok = _chart.SaveToStream(stream);

            //// Export the data from stream to byte array for sending. 
            //byte[] data = new byte[stream.Length];
            //stream.Position = 0; // Stream must be rewinded to start, as the position is at the end after writing. 
            //stream.Read(data, 0, data.Length);
            //stream.Dispose();

        }

        ///// <summary>
        /////  Communication channel received new setting parameter. 
        /////  <para>Parses the setting value and controls chart accordingly. </para>
        ///// </summary>
        ///// <param name="sender"></param>
        ///// <param name="setting"></param>
        //private void _comm_NewSetting(object sender, PipeConnector.SettingsEventArgs setting)
        //{
        //    if (_chart == null)
        //        return;

        //    _LCDispatcher.BeginInvoke(new Action(() =>
        //    {
        //        switch (setting.Name.ToLower())
        //        {
        //            case "width":
        //                _chart.Width = float.Parse(setting.Value);
        //                break;
        //            case "height":
        //                _chart.Height = float.Parse(setting.Value);
        //                break;
        //        };
        //    }));
        //}

        ///// <summary>
        ///// New request received from client. 
        ///// <para>
        /////  Parse the request and act accordingly. 
        ///// </para>
        ///// </summary>
        ///// <param name="sender"></param>
        ///// <param name="req"></param>
        //private void _comm_NewReq(object sender, PipeConnector.ReqEventArgs req)
        //{
        //    _LCDispatcher.BeginInvoke(new Action(() =>
        //    {
        //        switch (req.ReqType)
        //        {
        //            case ReqPackage.ReqType.NewFrame:
        //                // new frame rendered. Render new frame. It'll be delivered to the client at new frame rendered event. 
        //                RenderNewFrame();
        //                break;
        //            case ReqPackage.ReqType.StartGen:
        //                // Requst to start generation of data. 
        //                CheckTimerInit();
        //                if (!_dataGenTimer.IsEnabled)
        //                {

        //                    CheckSeriesInit();
        //                    _dataGenTimer.Start();
        //                    _timeLastGen = _dataStartTime = Environment.TickCount;
        //                }
        //                break;
        //            case ReqPackage.ReqType.StopGen:
        //                // Request to stop data generation. 
        //                if (_dataGenTimer != null && _dataGenTimer.IsEnabled)
        //                {
        //                    _dataGenTimer.Stop();
        //                }
        //                break;
        //            case ReqPackage.ReqType.StartDebugger:
        //                /// request to lauch debugger. This is usefull for debugging service. 
        //                Debugger.Launch();
        //                break;
        //        }
        //    }));
        //}

        /// <summary>
        /// Renders new frame. 
        /// </summary>
        /// <remarks>
        /// <para>
        /// Later there might be some other means to force chart to render new frames. It's just not usually something that is reuired, as chart renders new frame when it needs to. After data or property changes and so on. 
        /// </para>
        /// </remarks>
        private Bitmap RenderNewFrame(double index)
        {
            if (_chart != null)
            {
                if (_chart.ViewXY.LineSeriesCursors[0].ValueAtXAxis == index) return null;
                if (_chart.ViewXY.XAxes[0].Minimum > index)
                    _chart.ViewXY.LineSeriesCursors[0].ValueAtXAxis = _chart.ViewXY.XAxes[0].Minimum;
                else if (index > _chart.ViewXY.XAxes[0].Maximum)
                    _chart.ViewXY.LineSeriesCursors[0].ValueAtXAxis = _chart.ViewXY.XAxes[0].Maximum;
                else
                    _chart.ViewXY.LineSeriesCursors[0].ValueAtXAxis = index;
                _chart.BeginUpdate();
                byte[] aData = _chart.CaptureToByteArray(out var width, out var height, out var pixelFormat);
                _chart.EndUpdate();
                Bitmap bitmap = new Bitmap(width, height, System.Drawing.Imaging.PixelFormat.Format32bppArgb);
                System.Drawing.Imaging.BitmapData bitmapData = bitmap.LockBits(
                    new System.Drawing.Rectangle(0, 0, width, height), System.Drawing.Imaging.ImageLockMode.ReadOnly,
                    System.Drawing.Imaging.PixelFormat.Format32bppArgb);
                IntPtr ipDst = bitmapData.Scan0;
                int iRowByteCount = width * 4;
                int iSrcIndex = 0;
                for (int iY = 0; iY < height; iY++)
                {
                    Marshal.Copy(aData, iSrcIndex, ipDst, iRowByteCount);
                    ipDst = new IntPtr(ipDst.ToInt64() + bitmapData.Stride);
                    iSrcIndex += iRowByteCount;
                }
                bitmap.UnlockBits(bitmapData);
                LOG($"new bitmap {index} for {_chart.Title.Text} created...");
                return bitmap;
            }

            return null;

        }

        private void UpdateCursorResult(LightningChartUltimate chart)
        {
            //Disable rendering, strongly recommended before updating chart properties
            chart.BeginUpdate();

            //Get cursor
            LineSeriesCursor cursor = chart.ViewXY.LineSeriesCursors[0];

            //Get annotation
            AnnotationXY cursorValueDisplay = chart.ViewXY.Annotations[0];

            //Set annotation target. The location is relative to target. 
            //Use graph bottom as target Y value. 
            var _minValX = chart.ViewXY.PointLineSeries.SelectMany(p => p.Points).Min(p => p.X);
            var _maxValX = chart.ViewXY.PointLineSeries.SelectMany(p => p.Points).Max(p => p.X);
            float targetYCoordinate = DpiHelper.DipToPx((float)chart.ViewXY.GetMarginsRect().Bottom);
            chart.ViewXY.YAxes[0].CoordToValue(targetYCoordinate, out var seriesYValue, false);
            cursorValueDisplay.TargetAxisValues.X = cursor.ValueAtXAxis;
            cursorValueDisplay.TargetAxisValues.Y = seriesYValue;
            StringBuilder sb = new StringBuilder();
            string resultStringFormat = "{1,-10}:{0,10:N3}";
            string value = string.Format(resultStringFormat, seriesYValue, chart.ViewXY.YAxes[0].Title.Text);
            sb.AppendLine(value);
            sb.AppendLine();
            sb.AppendLine($"Time: {cursor.ValueAtXAxis:0.0##} {chart.ViewXY.XAxes[0].Units.Text}");
          
            //Set text
            cursorValueDisplay.Text = sb.ToString();
            var middleX = (_maxValX + _minValX) / 2;

            if (cursor.ValueAtXAxis > middleX)
            {
                if (cursorValueDisplay.LocationRelativeOffset.X > 0)
                {
                    cursorValueDisplay.LocationRelativeOffset.X *= -1.0F;
                }
            }
            else if (cursor.ValueAtXAxis <= middleX)
            {
                if (cursorValueDisplay.LocationRelativeOffset.X < 0)
                {
                    cursorValueDisplay.LocationRelativeOffset.X *= -1.0F;
                }
            }

            cursorValueDisplay.Visible = true;

            //Allow chart rendering
            chart.EndUpdate();
        }
        public void Dispose()
        {
            DispatcherOperation op = Application.Current.Dispatcher.BeginInvoke(new Action(() =>
            {
                LOG($"Called Dispose method for {_chart.Title.Text}");
                if (_chart != null)
                {
                    _chart.Dispose();
                    _chart = null;
                }
            }));
        }
    }
}